#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Model that represents a common RGB color and provides 
simple conversions to other common color systems.
"""

__all__ = ["Color", "ColorCompatibleT"]

import string
import typing

from hikari.internal_utilities import assertions


class Color(int, typing.SupportsInt):
    """Representation of a color.
    
    This value is immutable.

    This is a specialization of :obj:`int` which provides alternative overrides for common methods and color system
    conversions.

    This currently supports:

    * RGB
    * RGB (float)
    * 3-digit hex codes (e.g. 0xF1A -- web safe)
    * 6-digit hex codes (e.g. 0xFF11AA)
    * 3-digit RGB strings (e.g. #1A2 -- web safe)
    * 6-digit RGB hash strings (e.g. #1A2B3C)

    Examples
    --------
    Examples of conversions to given formats include:
    .. code-block::python

        >>> c = Color(0xFF051A)
        Color(r=0xff, g=0x5, b=0x1a)
        >>> hex(c)
        0xff051a
        >>> c.hex_code
        #FF051A
        >>> str(c)
        #FF051A
        >>> int(c)
        16712986
        >>> c.rgb
        (255, 5, 26)
        >>> c.rgb_float
        (1.0, 0.0196078431372549, 0.10196078431372549)

    Alternatively, if you have an arbitrary input in one of the above formats that you wish to become a color, you can
    use the get-attribute operator on the class itself to automatically attempt to resolve the color:
    .. code-block::python

        >>> Color[0xFF051A]
        Color(r=0xff, g=0x5, b=0x1a)
        >>> Color[16712986]
        Color(r=0xff, g=0x5, b=0x1a)
        >>> c = Color[(255, 5, 26)]
        Color(r=0xff, g=0x5, b=1xa)
        >>> c = Color[[0xFF, 0x5, 0x1a]]
        Color(r=0xff, g=0x5, b=1xa)
        >>> c = Color["#1a2b3c"]
        Color(r=0x1a, g=0x2b, b=0x3c)
        >>> c = Color["#1AB"]
        Color(r=0x11, g=0xaa, b=0xbb)
        >>> c = Color[(1.0, 0.0196078431372549, 0.10196078431372549)]
        Color(r=0xff, g=0x5, b=0x1a)
        >>> c = Color[[1.0, 0.0196078431372549, 0.10196078431372549]]
        Color(r=0xff, g=0x5, b=0x1a)

    Examples of initialization of Color objects from given formats include:
    .. code-block::python

        >>> c = Color(16712986)
        Color(r=0xff, g=0x5, b=0x1a)
        >>> c = Color.from_rgb(255, 5, 26)
        Color(r=0xff, g=0x5, b=1xa)
        >>> c = Color.from_hex_code("#1a2b3c")
        Color(r=0x1a, g=0x2b, b=0x3c)
        >>> c = Color.from_hex_code("#1AB")
        Color(r=0x11, g=0xaa, b=0xbb)
        >>> c = Color.from_rgb_float(1.0, 0.0196078431372549, 0.10196078431372549)
        Color(r=0xff, g=0x5, b=0x1a)
    """

    __slots__ = ()

    def __new__(cls, raw_rgb: typing.SupportsInt) -> "Color":
        assertions.assert_in_range(raw_rgb, 0, 0xFF_FF_FF, "integer value")
        return super(Color, cls).__new__(cls, raw_rgb)

    def __repr__(self) -> str:
        r, g, b = self.rgb
        return f"Color(r={hex(r)}, g={hex(g)}, b={hex(b)})"

    def __str__(self) -> str:
        return self.hex_code

    def __int__(self):
        # Binary-OR to make raw int.
        return self | 0

    @property
    def rgb(self) -> typing.Tuple[int, int, int]:
        """The RGB representation of this Color. Represented a tuple of R, G, B. Each value is in the 
        range [0, 0xFF]."""
        return (self >> 16) & 0xFF, (self >> 8) & 0xFF, self & 0xFF

    @property
    def rgb_float(self) -> typing.Tuple[float, float, float]:
        """Return the floating-point RGB representation of this Color. Represented as a tuple of R, G, B. 
        Each value is in the range [0, 1]."""
        r, g, b = self.rgb
        return r / 0xFF, g / 0xFF, b / 0xFF

    @property
    def hex_code(self) -> str:
        """The six-digit hexadecimal color code for this Color. This is prepended with a ``#`` symbol, 
        and will bein upper case.

        Example
        -------
        ``#1A2B3C``
        """
        return "#" + self.raw_hex_code

    @property
    def raw_hex_code(self) -> str:
        """The raw hex code.

        Example
        -------
        ``1A2B3C``
        """
        components = self.rgb
        return "".join(hex(c)[2:].zfill(2) for c in components).upper()

    @property
    def is_web_safe(self) -> bool:
        """``True`` if this color is a web-safe color, ``False`` otherwise."""
        hex_code = self.raw_hex_code
        return all(_all_same(*c) for c in (hex_code[:2], hex_code[2:4], hex_code[4:]))

    @classmethod
    def from_rgb(cls, red: int, green: int, blue: int) -> "Color":
        """Convert the given RGB colorspace represented in values within the range [0, 255]: [0x0, 0xFF], 
        to a :obj:`Color` object.

        Parameters
        ----------
        red : :obj:`int`
            Red channel.
        green : :obj:`int`
            Green channel.
        blue : :obj:`int`
            Blue channel.

        Returns
        -------
        :obj:`Color`
            A Color object.

        Raises
        ------
        :obj:`ValueError`
            If red, green, or blue are outside the range [0x0, 0xFF].
        """
        assertions.assert_in_range(red, 0, 0xFF, "red")
        assertions.assert_in_range(green, 0, 0xFF, "green")
        assertions.assert_in_range(blue, 0, 0xFF, "blue")
        # noinspection PyTypeChecker
        return cls((red << 16) | (green << 8) | blue)

    @classmethod
    def from_rgb_float(cls, red_f: float, green_f: float, blue_f: float) -> "Color":
        """Convert the given RGB colorspace represented using floats in 
        the range [0, 1] to a :obj:`Color` object.

        Parameters
        ----------
        red_f : :obj:`float`
            Red channel.
        green_f : :obj:`float`
            Green channel.
        blue_f : :obj:`float`
            Blue channel.

        Returns
        -------
        :obj:`Color`
            A Color object.

        Raises
        ------
        :obj:`ValueError`
            If red, green or blue are outside the range [0, 1].
        """
        assertions.assert_in_range(red_f, 0, 1, "red")
        assertions.assert_in_range(green_f, 0, 1, "green")
        assertions.assert_in_range(blue_f, 0, 1, "blue")
        # noinspection PyTypeChecker
        return cls.from_rgb(int(red_f * 0xFF), int(green_f * 0xFF), int(blue_f * 0xFF))

    @classmethod
    def from_hex_code(cls, hex_code: str) -> "Color":
        """Consumes a string hexadecimal color code and produces a :obj:`Color`.

        The inputs may be of the following format (case insensitive):
        ``1a2``, ``#1a2``, ``0x1a2`` (for websafe colors), or
        ``1a2b3c``, ``#1a2b3c``, ``0x1a2b3c`` (for regular 3-byte color-codes).

        Parameters
        ----------
        hex_code : :obj:`str`
            A hexadecimal color code to parse.

        Returns
        -------
        :obj:`Color`
            A corresponding Color object.

        Raises
        ------
        :obj:`ValueError`
            If ``hex_code`` is not a hexadecimal or is a inalid length.
        """
        if hex_code.startswith("#"):
            hex_code = hex_code[1:]
        elif hex_code.startswith(("0x", "0X")):
            hex_code = hex_code[2:]

        if not all(c in string.hexdigits for c in hex_code):
            raise ValueError("Color code must be hexadecimal")

        if len(hex_code) == 3:
            # Web-safe
            components = (int(c, 16) for c in hex_code)
            # noinspection PyTypeChecker
            return cls.from_rgb(*[(c << 4 | c) for c in components])

        if len(hex_code) == 6:
            return cls.from_rgb(int(hex_code[:2], 16), int(hex_code[2:4], 16), int(hex_code[4:6], 16))

        raise ValueError("Color code is invalid length. Must be 3 or 6 digits")

    @classmethod
    def from_int(cls, i: typing.SupportsInt) -> "Color":
        """Create a color from a raw integer that Discord can understand.

        Parameters
        ----------
        i : :obj:`typing.SupportsInt`
            The raw color integer.

        Returns
        -------
        :obj:`Color`
            The Color object.
        """
        return cls(i)

    # Partially chose to override these as the docstrings contain typos according to Sphinx.
    @classmethod
    def from_bytes(cls, bytes_: typing.Sequence[int], byteorder: str, *, signed: bool = True) -> "Color":
        """Converts the color from bytes."""
        return Color(super().from_bytes(bytes_, byteorder, signed=signed))

    def to_bytes(self, length: int, byteorder: str, *, signed: bool = True) -> bytes:
        """Converts the color code to bytes."""
        return super().to_bytes(length, byteorder, signed=signed)

    @classmethod
    def __class_getitem__(cls, color: "ColorCompatibleT") -> "Color":
        if isinstance(color, cls):
            return color
        elif isinstance(color, int):
            return cls.from_int(color)
        elif isinstance(color, (list, tuple)):
            assertions.assert_that(
                len(color) == 3, f"color must be an RGB triplet if set to a {type(color).__name__} type"
            )

            if _all_same(*map(type, color)):
                first = color[0]
                if isinstance(first, float):
                    return cls.from_rgb_float(*color)
                if isinstance(first, int):
                    return cls.from_rgb(*color)

                raise ValueError(
                    "all three channels must be all int or all float types if setting the color to an RGB triplet"
                )
        if isinstance(color, str):
            is_start_hash_or_hex_literal = color.casefold().startswith(("#", "0x"))
            is_hex_digits = all(c in string.hexdigits for c in color) and len(color) in (3, 6)
            if is_start_hash_or_hex_literal or is_hex_digits:
                return cls.from_hex_code(color)

        raise ValueError(f"Could not transform {color!r} into a {cls.__qualname__} object")


def _all_same(first, *rest):
    for r in rest:
        if r != first:
            return False

    return True


#: Any type that can be converted into a color object.
ColorCompatibleT = typing.Union[
    Color,
    typing.SupportsInt,
    typing.Tuple[typing.SupportsInt, typing.SupportsInt, typing.SupportsInt],
    typing.Tuple[typing.SupportsFloat, typing.SupportsFloat, typing.SupportsFloat],
    typing.Sequence[typing.SupportsInt],
    typing.Sequence[typing.SupportsFloat],
    str,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Basic transformation utilities."""
__all__ = [
    "CastInputT",
    "CastOutputT",
    "DefaultT",
    "TypeCastT",
    "ResultT",
    "nullable_cast",
    "try_cast",
    "put_if_specified",
    "image_bytes_to_image_data",
]

import base64
import contextlib
import typing

CastInputT = typing.TypeVar("CastInputT")
CastOutputT = typing.TypeVar("CastOutputT")
DefaultT = typing.TypeVar("DefaultT")
TypeCastT = typing.Callable[[CastInputT], CastOutputT]
ResultT = typing.Union[CastOutputT, DefaultT]


def nullable_cast(value: CastInputT, cast: TypeCastT) -> ResultT:
    """
    Attempts to cast the given ``value`` with the given ``cast``, but only if the 
    ``value`` is not ``None``. If it is ``None``, then ``None`` is returned instead.
    """
    if value is None:
        return None
    return cast(value)


def try_cast(value: CastInputT, cast: TypeCastT, default: DefaultT = None) -> ResultT:
    """Try to cast the given value to the given cast.
    
    If it throws a :obj:`Exception` or derivative, it will return ``default`` instead 
    of the cast value instead.
    """
    with contextlib.suppress(Exception):
        return cast(value)
    return default


def put_if_specified(
    mapping: typing.Dict[typing.Hashable, typing.Any],
    key: typing.Hashable,
    value: typing.Any,
    type_after: typing.Optional[TypeCastT] = None,
) -> None:
    """Add a value to the mapping under the given key as long as the value is not :obj:`typing.Literal`

    Parameters
    ----------
    mapping : :obj:`typing.Dict` [ :obj:`typing.Hashable`, :obj:`typing.Any` ]
        The mapping to add to.
    key : :obj:`typing.Hashable`
        The key to add the value under.
    value : :obj:`typing.Any`
        The value to add.
    type_after : :obj:`TypeCastT`, optional
        Type to apply to the value when added.
    """
    if value is not ...:
        if type_after:
            mapping[key] = type_after(value)
        else:
            mapping[key] = value


def image_bytes_to_image_data(img_bytes: typing.Optional[bytes] = None) -> typing.Optional[str]:
    """Encode image bytes into an image data string.

    Parameters
    ----------
    img_bytes : :obj:`bytes`, optional
        The image bytes.

    Raises
    ------
    :obj:`ValueError`
        If the image type passed is not supported.

    Returns
    -------
    :obj:`str`, optional
        The ``image_bytes`` given encoded into an image data string or ``None``.

    Note
    ----
    Supported image types: ``.png``, ``.jpeg``, ``.jfif``, ``.gif``, ``.webp``
    """
    if img_bytes is None:
        return None

    if img_bytes[:8] == b"\211PNG\r\n\032\n":
        img_type = "image/png"
    elif img_bytes[6:10] in (b"Exif", b"JFIF"):
        img_type = "image/jpeg"
    elif img_bytes[:6] in (b"GIF87a", b"GIF89a"):
        img_type = "image/gif"
    elif img_bytes.startswith(b"RIFF") and img_bytes[8:12] == b"WEBP":
        img_type = "image/webp"
    else:
        raise ValueError("Unsupported image type passed")

    image_data = base64.b64encode(img_bytes).decode()

    return f"data:{img_type};base64,{image_data}"


def try_cast_or_defer_unary_operator(type_):
    return lambda data: try_cast(data, type_, data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Date/Time utilities."""
__all__ = ["parse_http_date", "parse_iso_8601_ts", "discord_epoch_to_datetime", "unix_epoch_to_ts"]

import datetime
import email.utils
import re


def parse_http_date(date_str: str) -> datetime.datetime:
    """Return the HTTP date as a datetime object.

    Parameters
    ----------
    date_str : :obj:`str`
        The RFC-2822 (section 3.3) compliant date string to parse.

    See also
    --------
    `<https://www.ietf.org/rfc/rfc2822.txt>`_
    """
    return email.utils.parsedate_to_datetime(date_str)


ISO_8601_DATE_PART = re.compile(r"^(\d{4})-(\d{2})-(\d{2})")
# We don't always consistently get the subsecond part, it would seem. I have only
# observed this in raw_guild_members_chunk, however...
ISO_8601_TIME_PART = re.compile(r"T(\d{2}):(\d{2}):(\d{2})(?:\.(\d{1,6}))?", re.I)
ISO_8601_TZ_PART = re.compile(r"([+-])(\d{2}):(\d{2})$")


def parse_iso_8601_ts(date_string: str) -> datetime.datetime:
    """Parses an ISO 8601 date string into a datetime object

    Parameters
    ----------
    date_string : :obj:`str`
        The ISO 8601 compliant date string to parse.

    See also
    --------
    `<https://en.wikipedia.org/wiki/ISO_8601>`_
    """
    year, month, day = map(int, ISO_8601_DATE_PART.findall(date_string)[0])

    time_part = ISO_8601_TIME_PART.findall(date_string)[0]
    hour, minute, second, partial = time_part

    # Pad the millisecond part if it is not in microseconds, otherwise Python will complain.
    partial = partial + (6 - len(partial)) * "0"
    hour, minute, second, partial = int(hour), int(minute), int(second), int(partial)
    if date_string.endswith(("Z", "z")):
        timezone = datetime.timezone.utc
    else:
        sign, tz_hour, tz_minute = ISO_8601_TZ_PART.findall(date_string)[0]
        tz_hour, tz_minute = int(tz_hour), int(tz_minute)
        offset = datetime.timedelta(hours=tz_hour, minutes=tz_minute)
        if sign == "-":
            offset = -offset
        timezone = datetime.timezone(offset)

    return datetime.datetime(year, month, day, hour, minute, second, partial, timezone)


#: This represents the 1st January 2015 as the number of seconds since 1st January 1970 (Discord epoch)
DISCORD_EPOCH = 1_420_070_400


def discord_epoch_to_datetime(epoch: int) -> datetime.datetime:
    """Parses a discord epoch into a datetime object

    Parameters
    ----------
    epoch : :obj:`int`
        Number of milliseconds since 1/1/2015 (UTC)

    Returns
    -------
    :obj:`datetime.datetime`
        Number of seconds since 1/1/1970 within a datetime object (UTC).
    """
    return datetime.datetime.fromtimestamp(epoch / 1000 + DISCORD_EPOCH, datetime.timezone.utc)


def unix_epoch_to_ts(epoch: int) -> datetime.datetime:
    """Parses a unix epoch to a datetime object

    Parameters
    ----------
    epoch : :obj:`int`
        Number of milliseconds since 1/1/1970 (UTC)

    Returns
    -------
    :obj:`datetime.datetime`
        Number of seconds since 1/1/1970 within a datetime object (UTC).
    """
    return datetime.datetime.fromtimestamp(epoch / 1000, datetime.timezone.utc)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe both custom and Unicode
emojis on Discord.
"""
import typing

from hikari.core import entities
from hikari.core import snowflakes
from hikari.core import users
from hikari.internal_utilities import marshaller

__all__ = ["Emoji", "UnicodeEmoji", "UnknownEmoji", "GuildEmoji"]


@marshaller.attrs(slots=True)
class Emoji(entities.HikariEntity, entities.Deserializable):
    """Base class for all emojis."""


@marshaller.attrs(slots=True)
class UnicodeEmoji(Emoji):
    """Represents a unicode emoji."""

    #: The codepoints that form the emoji.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)


@marshaller.attrs(slots=True)
class UnknownEmoji(Emoji, snowflakes.UniqueEntity):
    """Represents a unknown emoji."""

    #: The name of the emoji.
    #:
    #: :type: :obj:`str`, optional
    name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None)

    #: Wheter the emoji is animated.
    #:
    #: :type: :obj:`bool`
    is_animated: bool = marshaller.attrib(raw_name="animated", deserializer=bool, if_undefined=False)


@marshaller.attrs(slots=True)
class GuildEmoji(UnknownEmoji):
    """Represents a guild emoji."""

    #: The whitelisted role IDs to use this emoji.
    #:
    #: :type: :obj:`typing.Set` [ :obj:`snowflakes.Snowflake` ]
    role_ids: typing.Set[snowflakes.Snowflake] = marshaller.attrib(
        raw_name="roles",
        deserializer=lambda roles: {snowflakes.Snowflake.deserialize(r) for r in roles},
        if_undefined=dict,
    )

    #: The user that created the emoji.
    #:
    #: Note
    #: ----
    #: This will be ``None`` if you are missing ``MANAGE_EMOJIS``` permission
    #: in the server the emoji is from
    #: (<https://github.com/discordapp/discord-api-docs/issues/593#issuecomment-386477863>)
    #:
    #: :type: :obj:`users.User`, optional
    user: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_none=None, if_undefined=None
    )

    #: Whether this emoji must be wrapped in colons.
    #:
    #: :type: :obj:`bool`, optional
    is_colons_required: typing.Optional[bool] = marshaller.attrib(
        raw_name="require_colons", deserializer=bool, if_undefined=None
    )

    #: Wheter the emoji is managed by an integration.
    #:
    #: :type: :obj:`bool`, optional
    is_managed: typing.Optional[bool] = marshaller.attrib(raw_name="managed", deserializer=bool, if_undefined=None)


def deserialize_reaction_emoji(payload: typing.Dict) -> typing.Union[UnicodeEmoji, UnknownEmoji]:
    """Deserialize a reaction emoji into an emoji."""
    if payload.get("id"):
        return UnknownEmoji.deserialize(payload)

    return UnicodeEmoji.deserialize(payload)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities related to discord's Oauth2 flow."""
__all__ = ["Application", "Owner", "OwnGuild", "Team", "TeamMember", "TeamMembershipState"]

import enum
import typing

from hikari.core import entities
from hikari.core import guilds
from hikari.core import permissions
from hikari.core import snowflakes
from hikari.core import users
from hikari.internal_utilities import cdn
from hikari.internal_utilities import marshaller


@marshaller.attrs(slots=True)
class OwnGuild(guilds.PartialGuild):
    """Represents a user bound partial guild object,
    returned by GET Current User Guilds.
    """

    #: Whether the current user owns this guild.
    #:
    #: :type: :obj:`bool`
    is_owner: bool = marshaller.attrib(raw_name="owner", deserializer=bool)

    #: The guild level permissions that apply to the current user or bot.
    #:
    #: :type: :obj:`permissions.Permission`
    my_permissions: permissions.Permission = marshaller.attrib(
        raw_name="permissions", deserializer=permissions.Permission
    )


@enum.unique
class TeamMembershipState(enum.IntEnum):
    """Represents the state of a user's team membership."""

    #: Denotes the user has been invited to the team but has yet to accept.
    INVITED = 1

    #: Denotes the user has accepted the invite and is now a member.
    ACCEPTED = 2


@marshaller.attrs(slots=True)
class TeamMember(entities.HikariEntity, entities.Deserializable):
    """Represents a member of a Team."""

    #: The state of this user's membership.
    #:
    #: :type: :obj:`TeamMembershipState`
    membership_state: TeamMembershipState = marshaller.attrib(deserializer=TeamMembershipState)

    #: This member's permissions within a team.
    #: Will always be ``["*"]`` until Discord starts using this.
    #:
    #: :type: :obj:`typing.Set` [ :obj:`str` ]
    permissions: typing.Set[str] = marshaller.attrib(deserializer=set)

    #: The ID of the team this member belongs to.
    #:
    #: :type: :obj:`snowflakes.Snowflake`
    team_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The user object of this team member.
    #:
    #: :type: :obj:`TeamUser`
    user: users.User = marshaller.attrib(deserializer=users.User.deserialize)


@marshaller.attrs(slots=True)
class Team(snowflakes.UniqueEntity, entities.Deserializable):
    """This represents a Team and it's members."""

    #: The hash of this team's icon, if set.
    #:
    #: :type: :obj:`str`, optional
    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str)

    #: The member's that belong to this team.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`snowflakes.Snowflake`, :obj:`TeamMember` ]
    members: typing.Mapping[snowflakes.Snowflake, TeamMember] = marshaller.attrib(
        deserializer=lambda members: {m.user.id: m for m in map(TeamMember.deserialize, members)}
    )

    #: The snowflake ID of this team's owner.
    #:
    #: :type: :obj:`snowflakes.Snowflake`
    owner_user_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    @property
    def icon_url(self) -> typing.Optional[str]:
        """The url of this team's icon, if set."""
        return self.format_icon_url()

    def format_icon_url(self, fmt: str = "png", size: int = 2048) -> typing.Optional[str]:
        """Generate the icon url for this team if set.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this url, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ``webp``.
        size : :obj:`int`
            The size to set for the url, defaults to ``2048``. Can be any power
            of two between 16 and 2048 inclusive.

        Returns
        -------
        :obj:`str`, optional
            The string url.
        """
        if self.icon_hash:
            return cdn.generate_cdn_url("team-icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None


@marshaller.attrs(slots=True)
class Owner(users.User):
    """Represents the user who owns an application, may be a team user."""

    #: This user's flags.
    #:
    #: :type: :obj:`int`
    flags: int = marshaller.attrib(deserializer=int)

    @property
    def is_team_user(self) -> bool:
        """If this user is a Team user (the owner of an application that's owned by a team)."""
        return bool((self.flags >> 10) & 1)


@marshaller.attrs(slots=True)
class Application(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents the information of an Oauth2 Application."""

    #: The name of this application.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: The description of this application, will be an empty string if unset.
    #:
    #: :type: :obj:`str`
    description: str = marshaller.attrib(deserializer=str)

    #: Whether the bot associated with this application is public.
    #: Will be ``None`` if this application doesn't have an associated bot.
    #:
    #: :type: :obj:`bool`, optional
    is_bot_public: typing.Optional[bool] = marshaller.attrib(
        raw_name="bot_public", deserializer=bool, if_undefined=None
    )

    #: Whether the bot associated with this application is requiring code grant
    #: for invites. Will be ``None`` if this application doesn't have a bot.
    #:
    #: :type: :obj:`bool`, optional
    is_bot_code_grant_required: typing.Optional[bool] = marshaller.attrib(
        raw_name="bot_require_code_grant", deserializer=bool, if_undefined=None
    )

    #: The object of this application's owner.
    #: This should always be ``None`` in application objects retrieved outside
    #: Discord's oauth2 flow.
    #:
    #: :type: :obj:`Owner`, optional
    owner: typing.Optional[Owner] = marshaller.attrib(deserializer=Owner.deserialize, if_undefined=None)

    #: A collection of this application's rpc origin urls, if rpc is enabled.
    #:
    #: :type: :obj:`typing.Set` [ :obj:`str` ], optional
    rpc_origins: typing.Optional[typing.Set[str]] = marshaller.attrib(deserializer=set, if_undefined=None)

    #: This summary for this application's primary SKU if it's sold on Discord.
    #: Will be an empty string if unset.
    #:
    #: :type: :obj:`str`
    summary: str = marshaller.attrib(deserializer=str)

    #: The base64 encoded key used for the GameSDK's ``GetTicket``.
    #:
    #: :type: :obj:`bytes`, optional
    verify_key: typing.Optional[bytes] = marshaller.attrib(
        deserializer=lambda key: bytes(key, "utf-8"), if_undefined=None
    )

    #: The hash of this application's icon if set.
    #:
    #: :type: :obj:`str`, optional
    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str, if_undefined=None)

    #: This application's team if it belongs to one.
    #:
    #: :type: :obj:`Team`, optional
    team: typing.Optional[Team] = marshaller.attrib(deserializer=Team.deserialize, if_undefined=None, if_none=None)

    #: The ID of the guild this application is linked to
    #: if it's sold on Discord.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    guild_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )

    #: The ID of the primary "Game SKU" of a game that's sold on Discord.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    primary_sku_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )

    #: The url slug that links to this application's store page
    #: if it's sold on Discord.
    #:
    #: :type: :obj:`str`, optional
    slug: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: The hash of this application's cover image on it's store, if set.
    #:
    #: :type: :obj:`str`, optional
    cover_image_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="cover_image", deserializer=str, if_undefined=None
    )

    @property
    def icon_url(self) -> typing.Optional[str]:
        """The url for this team's icon, if set."""
        return self.format_icon_url()

    def format_icon_url(self, fmt: str = "png", size: int = 2048) -> typing.Optional[str]:
        """Generate the icon url for this application if set.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this url, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ```webp``.
        size : :obj:`int`
            The size to set for the url, defaults to ``2048``.
            Can be any power of two between 16 and 2048.

        Returns
        -------
        :obj:`str`, optional
            The string url.
        """
        if self.icon_hash:
            return cdn.generate_cdn_url("app-icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None

    @property
    def cover_image_url(self) -> typing.Optional[str]:
        """The url for this icon's store cover image, if set."""
        return self.format_cover_image_url()

    def format_cover_image_url(self, fmt: str = "png", size: int = 2048) -> typing.Optional[str]:
        """Generate the url for this application's store page's cover image is
        set and applicable.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this url, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ``webp``.
        size : :obj:`int`
            The size to set for the url, defaults to ``2048``.
            Can be any power of two between 16 and 2048.

        Returns
        -------
        :obj:`str`, optional
            The string url.
        """
        if self.cover_image_hash:
            return cdn.generate_cdn_url("app-assets", str(self.id), self.cover_image_hash, fmt=fmt, size=size)
        return None
